package Kumera::Base;

####
##
## This code is distributed under the Open Source Artistic Licence
## http://www.opensource.org/licenses/artistic-license.html
##
## Kumera copyright (c) 2002, CyberCyberCyberCyber Pty Ltd.
## http://www.cyber4.org/kumera/index.html
##
## $Log: Base.pm,v $
## Revision 1.3  2002/10/14 15:31:27  grumpy
## Made use of $Kumera::BaseDir with respect to relative configuration parameters
##
## Revision 1.2  2002/10/13 17:15:48  grumpy
##
## FLuff, just a check.
##
## Revision 1.1.1.1  2002/10/13 17:11:23  grumpy
## Added copyright.
##
##
####

require Exporter;
require Kumera;
@ISA = qw/Exporter Kumera/;

# more stuff for config
use XML::DOM;
use File::Basename;

sub new {

    # build object references
    my $self = shift;
    my $type = ref($self) || $self;

    # organise arguments
    my $args = shift || {};

    # make sure that args is a hash, it will fail if not
    my $this = { %{ $args }};

    $this = bless $this, $type;

    $parser = new XML::DOM::Parser;
    my $DOMTree = $parser->parsefile($Kumera::XMLConfig);

    # now go through the DOMTree and extract into a hash
    # the information we really need
    $this->{ CONFIG_DATA } = _parse_config_document($DOMTree);

    # now do the acl xml file
    $DOMTree = $parser->parsefile($this->getLoginFilename('acl'));

    # now go through the DOMTree and extract into a hash
    # the information we really need
    my $acl_data = _parse_acl_document($DOMTree);
    $this->{ USER_DATA } = $acl_data->[0];
    $this->{ PATH_DATA } = $acl_data->[1];

    if (defined $this->{ PATH_INFO }) {

        # build a correct path_info now
        $this->{ PATH_INFO } = $this->_build_path_info;

        $this->_debug('this->{ PATH_INFO }', $this->{ PATH_INFO });
    }

    # all these depend on a user_id being set, for lock files and so on
    if (defined $this->{ USER_ID }) {

        # all of these will check the hasPermission to see if
        # the uer has edit.section, and the file is locked out
        # by someone else, the order of these should not by
        # changed as lower ones rely on higher one (in theory
        # but cacheing of temp.results are done, esp. for locked_out_by)

        # and now edit_filename
        $this->{ EDIT_FILENAME } = $this->_build_edit_filename;

        # and now lock_filename
        $this->{ LOCK_FILENAME } = $this->_build_lock_filename;

        # and then html_filename
        $this->{ HTML_FILENAME } = $this->_build_html_filename;

        # and then cache this
        $this->{ LOCKED_OUT_BY } = $this->locked_out_by;

        # and then a browse_filename (must go last)
        $this->{ BROWSE_FILENAME } = $this->_build_browse_filename;

        $this->_debug('this->{ EDIT_FILENAME }', $this->{ EDIT_FILENAME });
        $this->_debug('this->{ LOCK_FILENAME }', $this->{ LOCK_FILENAME });
        $this->_debug('this->{ HTML_FILENAME }', $this->{ HTML_FILENAME });

        $this->_debug('this->{ BROWSE_FILENAME }', $this->{ BROWSE_FILENAME });
    }


    return $this;
}

sub _parse_config_document {

    my $DOMTree = shift;
    my $document;

    Kumera::_debug('DOMTree->getNodeName()', $DOMTree->getNodeName());

    # there may, or may not be a DOCTYPE, which we ignore
    my $docNodes = $DOMTree->getChildNodes();
    for (my $i = 0 ; $i < $docNodes->getLength() ; $i++ ) {

        my $docNode = $docNodes->item($i);

        # look for the config element
        if ($docNode->getNodeName() eq 'Kumera:config') {

            # parse the config section, should only ever
            # be once config section ...
            $document = _parse_config($docNode);

            Kumera::_debug('document', $document);

        }
    }

    # and return a reference to the hash
    return $document;
}

sub _parse_config {

    my $configTree = shift;
    my %configHash;
    my %parameterHash;

    Kumera::_debug('configTree->getNodeName()', $configTree->getNodeName());

    # config is made up of a data section
    my $configNodes = $configTree->getChildNodes();
    for (my $i = 0 ; $i < $configNodes->getLength() ; $i++ ) {
        my $configNode = $configNodes->item($i);

        # check if it is a data node first
        if ($configNode->getNodeName() eq 'process') {

            my $name = $configNode->getAttribute('name');

            Kumera::_debug('process-name', $name);

            # parse the data section and save it under the process name
            $configHash{ $name } = _parse_process($configNode);

            Kumera::_debug('configHash{'.$name.'}', $configHash{ $name });

        } elsif ($configNode->getNodeName() eq 'parameter') {

            my $name = $configNode->getAttribute('name');
            my $value = $configNode->getFirstChild()->getData();

            Kumera::_debug('parameter-name', $name);
            Kumera::_debug('parameter-value', $value);

            $parameterHash{ $name } = $value;
        }
    }

    return [\%configHash, \%parameterHash];
}

sub _parse_process {

    my $processTree = shift;
    my %processHash;

    # get the attrbutes, as we need to pester it about the dir
    my $dir = $processTree->getAttribute('dir');

    # and remove any trailing '/' it might have
    chop $dir if $dir =~ m|/$|;

    Kumera::_debug('dir', $dir);

    Kumera::_debug('processTree->getNodeName()', $processTree->getNodeName());

    # process is made up of file at the moment
    my $processNodes = $processTree->getChildNodes();
    for (my $i = 0 ; $i < $processNodes->getLength() ; $i++ ) {
        my $processNode = $processNodes->item($i);

        # we need to extract the file name and file attributes
        Kumera::_debug('processNode->getNodeName()', $processNode->getNodeName());

        # check which process we are in first
        if ($processNode->getNodeName() eq 'file') {

            my $name = $processNode->getAttribute('name');
            my $relative = $processNode->getAttribute('type') eq 'relative';

            my $file = $processNode->getFirstChild()->getData();

            # check if file is absolute or relative
            $file = $Kumera::BaseDir.$dir.'/'.$file if $relative;

            Kumera::_debug(($relative ? 'relative' : 'absolute').' file', $file);

            # and put into the hash
            $processHash{ $name } = $file;
        }

    }

    return \%processHash;
}

# parse the acl xml file
sub _parse_acl_document {

    my $DOMTree = shift;
    my $document;

    Kumera::_debug('DOMTree->getNodeName()', $DOMTree->getNodeName());

    # there may, or may not be a DOCTYPE, which we ignore
    my $docNodes = $DOMTree->getChildNodes();
    for (my $i = 0 ; $i < $docNodes->getLength() ; $i++ ) {

        my $docNode = $docNodes->item($i);

        # look for the config element
        if ($docNode->getNodeName() eq 'Kumera:acl') {

            # parse the config section, add to documentHash
            $document = _parse_acl($docNode);

        }
    }

    # and return a reference to the hash
    return $document;
}

sub _parse_acl {

    my $aclTree = shift;
    my %userHash = ();
    my %roleHash = ();
    my %sectionHash = ();

    Kumera::_debug('aclTree->getNodeName()', $aclTree->getNodeName());

    # user is made up of user sections
    my $aclNodes = $aclTree->getChildNodes();
    for (my $i = 0 ; $i < $aclNodes->getLength() ; $i++ ) {

        my $aclNode = $aclNodes->item($i);

        # check if it is a data node first
        if ($aclNode->getNodeName() eq 'user') {

            my $id = $aclNode->getAttribute('id');

            # and now parse the content of the user element
            my $name = _parse_user_child('name', $aclNode);
            my $password = _parse_user_child('password', $aclNode);
            my $email = _parse_user_child('email', $aclNode);

            Kumera::_debug('id', $id);
            Kumera::_debug('password', $password);
            Kumera::_debug('name', $name);
            Kumera::_debug('email', $email);

            # put the name and password into the hash as
            # an array of two element
            my %data = ();

            $data{ NAME } = $name;
            $data{ PASSWORD } = $password;
            $data{ EMAIL } = $email;

            $userHash{ $id } = \%data;

            Kumera::_debug('userHash{'.$id.'}', $userHash{ $id });

        } elsif ($aclNode->getNodeName() eq 'role') {

            my $id = $aclNode->getAttribute('id');

            # and now parse the content of the user role
            # it needs what we have already for inheritence to work
            Kumera::_debug('role_id', $id);
            %roleHash = _parse_role_child(\%roleHash, $id, $aclNode);

        } elsif ($aclNode->getNodeName() eq 'section') {

            my $id = $aclNode->getAttribute('id');
            my $dir = $aclNode->getAttribute('dir');

            Kumera::_debug('section_id', $id);
            Kumera::_debug('dir', $dir);

            # and now parse the path children, using the information we already have
            %sectionHash = (%sectionHash, _parse_section_child(\%userHash, \%roleHash, $id, $dir, $aclNode));
        }
    }

    # the calling subroutine will deal with these
    return [ \%userHash, \%sectionHash ];
}

sub _parse_user_child {

    my $name = shift;
    my $userTree = shift;

    # loop through looking for this one
    my $userNodes = $userTree->getChildNodes();
    for (my $i = 0 ; $i < $userNodes->getLength() ; $i++ ) {

        my $userNode = $userNodes->item($i);

        if ($userNode->getNodeName() eq $name) {

            # found what we want, so return it (data in child of element)
            return $userNode->getFirstChild()->getData();
        }
    }

    return '';
}

sub _parse_role_child {

    my $roleHash = shift;
    my $role_id = shift;
    my $roleTree = shift;

    # loop through the child nodes
    my $roleNodes = $roleTree->getChildNodes();
    for (my $i = 0 ; $i < $roleNodes->getLength() ; $i++ ) {

        # grab the child
        my $roleNode = $roleNodes->item($i);

        # if it is an inheritence one, then grab the parent
        if ($roleNode->getNodeName() eq 'inherits') {

            my $parent_id = $roleNode->getFirstChild->getData();
            Kumera::_debug('parent_id', $parent_id);

            # do the inheritence
            Kumera::_debug('roleHash->{'.$role_id.'}', join ',', keys %{ $roleHash->{ $role_id }});
            Kumera::_debug('roleHash->{'.$parent_id.'}', join ',', keys %{ $roleHash->{ $parent_id }});

            my %tempRoleHash = ( %{ $roleHash->{ $parent_id }}, %{ $roleHash->{ $role_id }} );
            $roleHash->{ $role_id } = \%tempRoleHash;

            Kumera::_debug('tempRoleHash', join ',', keys %tempRoleHash);
            Kumera::_debug('roleHash->{'.$role_id.'}', join ',', keys %{ $roleHash->{ $role_id }});

        } elsif ($roleNode->getNodeName() eq 'permission') {

            my $id = $roleNode->getAttribute('id');
            Kumera::_debug('id', $id);

            $roleHash->{ $role_id }->{ $id } = 1;
            Kumera::_debug('roleHash->{'.$role_id.'}->{'.$id.'}', $roleHash->{ $role_id }->{ $id });
            Kumera::_debug('roleHash->{'.$role_id.'}', join ',', keys %{ $roleHash->{ $role_id }});
        }

    }

    return %{ $roleHash };
}

sub _parse_section_child {

    my $userHash = shift;
    my $roleHash = shift;
    my $section_id = shift;
    my $dir = shift;
    my $sectionTree = shift;

    # index by dir for quick access, rather then
    # by section_id which is used only for commit stuff
    my %sectionHash = ( $dir => { SECTION_ID => $section_id } );

    # loop through looking for this one
    my $sectionNodes = $sectionTree->getChildNodes();
    for (my $i = 0 ; $i < $sectionNodes->getLength() ; $i++ ) {

        my $sectionNode = $sectionNodes->item($i);

        if ($sectionNode->getNodeName() eq 'role') {

            # grab the role_id for this section
            my $role_id = $sectionNode->getAttribute('id');
            Kumera::_debug('role_id', $role_id);

            # now loop through and get the users under this role
            my $roleNodes = $sectionNode->getChildNodes();
            for (my $j = 0 ; $j < $roleNodes->getLength() ; $j++ ) {

                my $roleNode = $roleNodes->item($j);

                if ($roleNode->getNodeName() eq 'user') {
                    my $user_id = $roleNode->getFirstChild()->getData();

                    Kumera::_debug('user_id', $user_id);

                    # and now put it all together
                    my %tempUserHash = ( %{ $sectionHash{ $dir } }, $user_id => $roleHash->{ $role_id } );
                    Kumera::_debug('tempUserHash{'.$user_id.'}', join ',', keys %{ $tempUserHash{ $user_id }});

                    $sectionHash{ $dir } = \%tempUserHash;

                    Kumera::_debug('sectionHash{'.$dir.'}->{'.$user_id.'}', join ',', keys %{ $sectionHash{ $dir }->{ $user_id }} );
                    Kumera::_debug('sectionHash{'.$dir.'}->{ SECTION_ID }', $sectionHash{ $dir }->{ SECTION_ID });
                }
            }
        }
    }

    return %sectionHash;
}


# first we have functions to see if we can edit the file
sub is_lockable {

    my $this = shift;
    my $args = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO };

    # cannot do anything to file if it's the index file
    return 0 if $this->isIndexFile($path_info);

    # it may be not locked out, or locked out by this user
    # or they make have edit.section priv's
    my $edit_filename = $this->_build_edit_filename({ USER_ID => $user_id, PATH_INFO => $path_info });
    my $lock_filename = $this->_build_lock_filename({ USER_ID => $user_id, PATH_INFO => $path_info });

    $this->_debug('edit_filename', $edit_filename);
    $this->_debug('lock_filename', $lock_filename);

    $this->_debug('this->hasPermission(edit.section)', $this->hasPermission('edit.section'));

    # check that they can edit, and that either of the two editable
    # files exist ...
    return $this->hasPermission('edit') && (-w $edit_filename || -w $lock_filename);
}

sub is_locked_by {

    my $this = shift;
    my $args = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO };
    my $index_override = defined $args->{ INDEX_OVERRIDE };

    # index file cannot be locked, and this will cause funny errors through system
    # so we have an overide option to allow us to break the rules
    return 0 if $this->isIndexFile($path_info) && !$index_override;

    # my $lock_filename = $this->_build_lock_filename({ USER_ID => $user_id });

    # do this here rather then call _build_lock_filename because it loops
    my $edit_filename = $this->_build_edit_filename({ USER_ID => $user_id, PATH_INFO => $path_info });
    my $section_id = $this->_get_section_id($path_info);
    my $lock_filename;
    ($lock_filename = $edit_filename) =~ s/\.xml$/.$user_id.$section_id.xml/;

    $this->_debug('is_locked_by', $lock_filename);

    return -w $lock_filename;
}

# don't cache the result of this, as path_info may change
sub locked_out_by {

    my $this = shift;
    my $args = shift;

    # don't do $user_id here, has no meaning
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO };
    my $index_override = defined $args->{ INDEX_OVERRIDE };

    $this->_debug('index_override', $index_override);
    $this->_debug('path_info', $path_info);

    # index file cannot be locked out, unless we want to override this
    return '' if $this->isIndexFile($path_info) && !$index_override;

    # see which user it might be
    foreach my $user_id ($this->allUsers) {

        # is it locked out?
        if ($this->is_locked_by({ USER_ID => $user_id, PATH_INFO => $path_info, INDEX_OVERRIDE => $index_override })) {

            $this->_debug('locked_out_by', $user_id);

            # store away if relevant
            $this->{ LOCKED_OUT_BY } = $user_id if $path_info eq $this->{ PATH_INFO };

            return $user_id;
        }
    }

    return '';
}

sub is_deleteable {

    my $this = shift;
    my $args = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO };

    $this->_debug('user_id', $user_id);
    $this->_debug('path_info(7)', $path_info);

    # can't be deleted if an index file
    return 0 if $this->isIndexFile($path_info);

    # the file will either be locked out, or not locked out
    # if it is not locked out, they need delete.* permission to delete it
    # if this user has it locked out they need delete.* permission to delete it
    # if it is locked out by someone else, they need delete.section permission to delete it

    # the file must also exist in the data directory (may have been deleted previously
    # but picked up by upload's listing
    my $edit_filename = $this->_build_edit_filename({ USER_ID => $user_id, PATH_INFO => $path_info });
    my $lock_filename = $this->_build_lock_filename({ USER_ID => $user_id, PATH_INFO => $path_info });

    $this->_debug('edit_filename', $edit_filename);
    $this->_debug('lock_filename', $lock_filename);

    # unless it exists as an xml edit file, we can't delete the sucker
    return 0 unless (-w $edit_filename || -w $lock_filename);

    my $locked_out_by = $this->locked_out_by({ USER_ID => $user_id, PATH_INFO => $path_info });

    $this->_debug('this->hasPermission(delete.section)', $this->hasPermission('delete.section'));
    $this->_debug('this->hasPermission(delete.user)', $this->hasPermission('delete.user'));
    $this->_debug('locked_out_by', $locked_out_by);

    return
        !$locked_out_by && $this->hasPermission('delete')
        ||
        $locked_out_by eq $user_id && $this->hasPermission('delete')
        ||
        $this->hasPermission('delete.section');
}

sub _build_path_info {

    my $this = shift;

    my $path_info = $this->{ PATH_INFO };

    # a path_info may not have been given
    if ($path_info) {

        # if it ends with an '/' or is a directory, and add default filename
        chop $path_info if $path_info =~ m|/$|;

        # add a / to the end if it is a directory, and add the default to it
        $path_info .= $this->getBrowseFilename('index') if -d $this->getBrowseFilename('root').$path_info;

    } else {

        # use the home page if path_info not sensible
        $path_info = $this->getBrowseFilename('default');

    }

    $this->_debug('path_info', $path_info);

    return $path_info;
}

sub getPathInfo {

    my $this = shift;

    return $this->{ PATH_INFO };
}

sub _build_html_filename {

    my $this = shift;
    my $args = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO };

    # and make html what we'd expect
    my $html_filename = $this->_build_lock_filename({ USER_ID => $user_id, PATH_INFO => $path_info });
    $html_filename =~ s/\.xml$/\.html/;

    $this->_debug('html_filename', $html_filename);

    return $html_filename;
}

sub _build_browse_filename {

    my $this = shift;
    my $args = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO };

    my $browse_filename;

    # see if anyone has this locked out at the moment
    my $lock_user_id = $this->locked_out_by({ USER_ID => $user_id, PATH_INFO => $path_info });

    # is this the index file, in which case we dig up index_filename
    if ($this->isIndexFile($path_info)) {

        $browse_filename = $this->_build_index_browse_filename({ USER_ID => $user_id, PATH_INFO => $path_info });

    # does this user have it locked out, and have permission to edit it
    } elsif ($lock_user_id eq $user_id && $this->hasPermission('edit.user', $path_info)) {

        $browse_filename = $this->_build_html_filename({ USER_ID => $user_id, PATH_INFO => $path_info });

    # does another user have it locked out, and the user has permission to edit it
    } elsif ($lock_user_id && $lock_user_id ne $user_id && $this->hasPermission('edit.section', $path_info)) {

        $browse_filename = $this->_build_html_filename({ USER_ID => $this->locked_out_by({ USER_ID => $user_id, PATH_INFO => $path_info }) });

    # no-one has it locked out, so use plain html file to browse with
    } else {

        $browse_filename = $this->getBrowseFilename('root').$path_info;
    }

    $this->_debug('browse_filename', $browse_filename);

    return $browse_filename;
}

sub _build_edit_filename {

    my $this = shift;
    my $args = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO };

    # now, process the information we have been given
    my $edit_filename = $this->getEditFilename('root').$path_info;
    $edit_filename =~ s/\.html$/.xml/;

    $this->_debug('edit_filename', $edit_filename);

    return $edit_filename;
}

sub _build_lock_filename {

    my $this = shift;
    my $args = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO };

    $this->_debug('user_id', $user_id);

    my $lock_filename = $this->_build_edit_filename({ USER_ID => $user_id, PATH_INFO => $path_info });
    my $section_id = $this->_get_section_id($path_info);

    # either
    # (1) it is locked out by this user or not locked out, so we go
    #     for the ordinary path_info.filename.user_id.section_id.xml
    #     type of lock_filename
    # (3) someone else has it locked out, and we have edit.section permission
    #     so we need to play around with the names a bit

    # not locked out
    if (-w $this->_build_edit_filename({ USER_ID => $user_id, PATH_INFO => $path_info })) {

        $lock_filename =~ s/\.xml$/.$user_id.$section_id.xml/;

    # is locked out, and user has permission
    } elsif ($this->hasPermission('edit.section')) {

        my $lock_user_id = $this->locked_out_by({ USER_ID => $user_id, PATH_INFO => $path_info });
        $this->_debug('lock_user_id(1)', $lock_user_id);
        $lock_filename =~ s/\.xml$/.$lock_user_id.$section_id.xml/;

    # is locked out, but by this user (edit.section user may already have edit.user
    # so have this second to allow fo edit.seciton to be higher priority
    } elsif ($this->hasPermission('edit.user')) {

        $lock_filename =~ s/\.xml$/.$user_id.$section_id.xml/;

    }

    $this->_debug('lock_filename', $lock_filename);

    return $lock_filename;
}

sub _build_commit_filename {

    my $this = shift;
    my $args = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $html_filename = $args->{ HTML_FILENAME } || $this->{ HTML_FILENAME };

    # need so strip out the (could be any) user_id and section_id from the given path_info
    my $commit_filename;
    my $edit_root = $this->getEditFilename('root');
    ($commit_filename = $html_filename) =~ s/^$edit_root(.*?)\.(.*?)\.(.*?)\.html$/$1.html/;

    return $commit_filename;
}

#
# This will return the full filename to the
# XML element pointed to by <file name="arg1" location="return1">
# It will include appending the base_directory to the return1
# filename in a sensible way
#

sub _getFilename {

    my $this = shift;
    my $process = shift;
    my $name = shift;

    # these are way to noisy
    # $this->_debug('this', $this);
    # $this->_debug('this->{ CONFIG_DATA }', $this->{ CONFIG_DATA });
    # $this->_debug('this->{ CONFIG_DATA }->[0]', $this->{ CONFIG_DATA }->[0]);
    # $this->_debug('this->{ CONFIG_DATA }->[0]->{'.$process.'}', $this->{ CONFIG_DATA }->[0]->{ $process });
    $this->_debug('this->{ CONFIG_DATA }->[0]->{'.$process.'}->{'.$name.'}', $this->{ BASE_DIR }.$this->{ CONFIG_DATA }->[0]->{ $process }->{ $name });

    return $this->{ CONFIG_DATA }->[0]->{ $process }->{ $name };
}

sub getLoginFilename {

    my $this = shift;
    my $name = shift;

    return $this->_getFilename('login', $name);
}

sub getBrowseFilename {

    my $this = shift;
    my $name = shift;

    return $this->_getFilename('browse', $name);
}

sub getCreateFilename {

    my $this = shift;
    my $name = shift;

    return $this->_getFilename('create', $name);
}

sub getDeleteFilename {

    my $this = shift;
    my $name = shift;

    return $this->_getFilename('delete', $name);
}

sub getEditFilename {

    my $this = shift;
    my $name = shift;

    return $this->_getFilename('edit', $name);
}

sub getCommitFilename {

    my $this = shift;
    my $name = shift;

    return $this->_getFilename('commit', $name);
}

sub getUploadFilename {

    my $this = shift;
    my $name = shift;

    return $this->_getFilename('upload', $name);
}

sub getPrintFilename {

    my $this = shift;
    my $name = shift;

    return $this->_getFilename('print', $name);
}

sub getParameter {

    my $this = shift;
    my $name = shift;

    return $this->{ CONFIG_DATA }->[1]->{ $name };
}

sub menu {

    my $this = shift;

    my $types = shift;
    my $path_info = shift;

    my $message = shift;

    # blank message if we need it, other wrap it into a row
    $message = defined $message ? '<tr bgcolor="#cc3333"><td colspan="8"><font color="#ffffff"><b>'.$message.'</b></font></td></tr>' : '';

    # check for browse
    my $browse_url = $types->{'browse'} && $this->hasPermission('browse') ? '<a href="'.$Kumera::script_path.'/browse.cgi'.$path_info.'">browse</a>' : 'browse';

    # check for create
    my $create_url = $types->{'create'} && $this->hasPermission('create') ? '<a href="'.$Kumera::script_path.'/create.cgi'.$path_info.'">create</a>' : 'create';

    # check what can be done here, edit
    my $edit_url = $types->{'edit'} && $this->hasPermission('edit') ? '<a href="'.$Kumera::script_path.'/edit.cgi'.$path_info.'">edit</a>' : 'edit';

    # check for delete
    my $delete_url = $types->{'delete'} && $this->hasPermission('delete') ? '<a href="'.$Kumera::script_path.'/delete.cgi'.$path_info.'">delete</a>' : 'delete';

    # check for commit
    my $commit_url = $types->{'commit'} && $this->hasPermission('commit') ? '<a href="'.$Kumera::script_path.'/commit.cgi'.$path_info.'">commit</a>' : 'commit';

    # check for upload
    my $upload_url = $types->{'upload'} && $this->hasPermission('upload') ? '<a href="'.$Kumera::script_path.'/upload.cgi'.$path_info.'">upload</a>' : 'upload';

    # check if we can logout, which should be always be able to do
    my $logout_url = $types->{'logout'} && $this->hasPermission('logout') ? '<a href="'.$Kumera::script_path.'/login.cgi">logout</a>' : 'logout';

    return << "__END_OF_MENU__";

<center>
<table>$message
<tr bgcolor="#ffffff">
<td><b>user</b> $this->{ USER_ID }</td>
<td>$browse_url</td>
<td>$create_url</td>
<td>$edit_url</td>
<td>$delete_url</td>
<td>$commit_url</td>
<td>$upload_url</td>
<td>$logout_url</td>
</tr>
</table>
</center>

__END_OF_MENU__

}


#
# User and acl stuff
#

# This will return the name to the caller

sub getName {

    my $this = shift;
    my $user_id = shift;

    return '' unless $this && $user_id;

    return $this->{ USER_DATA }->{ $user_id }->{ NAME };
}

sub getEmail {

    my $this = shift;
    my $user_id = shift;

    return '' unless $this && $user_id;

    return $this->{ USER_DATA }->{ $user_id }->{ EMAIL };
}

sub validate {

    my $this = shift;
    my $user_id = shift;
    my $password = shift;

    $this->_debug('this', $this);
    $this->_debug('this->{ USER_DATA }', $this->{ USER_DATA });
    $this->_debug('this->{ USER_DATA }->{'.$user_id.'}', $this->{ USER_DATA }->{ $user_id });
    $this->_debug('this->{ USER_DATA }->{'.$user_id.'}->{ PASSWORD }', $this->{ USER_DATA }->{ $user_id }{ PASSWORD });

    return $password && $this->{ USER_DATA }->{ $user_id }{ PASSWORD } && $this->{ USER_DATA }->{ $user_id }->{ PASSWORD } eq $password;
}

sub allUsers {

    my $this = shift;

    return keys %{ $this->{ USER_DATA } };
}

sub _get_section_id {

    my $this = shift;
    my $path_info = shift;

    my $user_id = $this->{ USER_ID };
    $path_info = $this->{ PATH_INFO } unless defined $path_info;

    # need to do a dirname loop over the path_info until we get to root
    while ($path_info ne '/' && $path_info ne '/') {

        $this->_debug('path_info', $path_info);
        $this->_debug('defined this->{ PATH_DATA }->{'.$path_info.'}', defined $this->{ PATH_DATA }->{ $path_info } ? 'true' : 'false');
        $this->_debug('defined this->{ PATH_DATA }->{'.$path_info.'}->{'.$user_id.'}', defined $this->{ PATH_DATA }->{ $path_info }->{ $user_id } ? 'true' : 'false');

        # jump out if we have found something for this path_info and user;
        last if (defined $this->{ PATH_DATA }->{ $path_info } && defined $this->{ PATH_DATA }->{ $path_info }->{ $user_id });

        $path_info = dirname($path_info);
    }

    $this->_debug('path_info', $path_info);
    $this->_debug('this->{ PATH_DATA }->{'.$path_info.'}->{ SECTION_ID }', $this->{ PATH_DATA }->{ $path_info }->{ SECTION_ID });

    return $this->{ PATH_DATA }->{ $path_info }->{ SECTION_ID };
}

sub hasPermission {

    my $this = shift;
    my $permission = shift;
    my $path_info = shift;

    my $user_id = $this->{ USER_ID } ;
    $path_info = $this->{ PATH_INFO } unless defined $path_info;

    $this->_debug('user_id', $user_id);
    $this->_debug('permission', $permission);

    # need to do a dirname loop over the path_info until we get to root
    while ($path_info ne '/' && $path_info ne '.') {

        $this->_debug('path_info', $path_info);
        $this->_debug('defined this->{ PATH_DATA }->{'.$path_info.'}', defined $this->{ PATH_DATA }->{ $path_info } ? 'true' : 'false');
        $this->_debug('defined this->{ PATH_DATA }->{'.$path_info.'}->{'.$user_id.'}', defined $this->{ PATH_DATA }->{ $path_info }->{ $user_id } ? 'true' : 'false');

        # jump out if we have found something for this path_info and user;
        last if (defined $this->{ PATH_DATA }->{ $path_info } && defined $this->{ PATH_DATA }->{ $path_info }->{ $user_id });

        $path_info = dirname($path_info);
    }

    $this->_debug('path_info', $path_info);

    # there may be a compound permission, eg commit.section and commit.user, edit.user and edit.section
    foreach my $key (keys %{ $this->{ PATH_DATA }->{ $path_info }->{ $user_id }}) {

        $this->_debug('key', $key);

        # true if first part matches, eg 'commit' and 'commit.user' are a match
        return 1 if $key =~ /^$permission/;
    }

    return 0;
}

####
##
## stuff for rebuilding index files
##
####

use IO::File;
use File::Copy;

sub _build_index_edit_filename {

    my $this = shift;
    my $args = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO }; # used for dirname

    # we'll use the bog standard index filename
    return $this->getEditFilename('root').(dirname($path_info) ne '/' ? dirname($path_info) : '').$this->getEditFilename('index');
}

sub _build_index_lock_filename {

    my $this = shift;
    my $args = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO }; # used for dirname

    $this->_debug('user_id', $user_id);

    ##
    # and index file exists as any of
    #
    #   index.xml
    #   index.user_id.section_id.xml
    #   index.section_id.xml
    #
    # and depending on their permission, determines which one they get to
    # use, see, write to and read from

    # first, using the original path_info, see if the edited file is locked out
    # by a user at this point in time (don't need index_override here as path_info should not
    # be for an index_file, and if it is then we don't worry about it)
    my $lock_user_id = $this->locked_out_by({ USER_ID => $user_id, PATH_INFO => $path_info });
    $this->_debug('lock_user_id', $lock_user_id);

    # now alter path_info to the index filename
    $path_info = (dirname($path_info) ne '/' ? dirname($path_info) : '').$this->getEditFilename('index');

    # and build first part of the index_lock_filename
    my $index_lock_filename = $this->_build_edit_filename({ USER_ID => $user_id, PATH_INFO => $path_info });
    my $section_id = $this->_get_section_id($path_info);

    # if they are a section editor, they can edit any standard file locked out by
    # another user, so we want to write back to the index file of that user
    # so this is editing a file that is not owned by this user, assuming the
    # file is locked out ...
    if ($this->hasPermission('edit.section') && $lock_user_id && $user_id ne $lock_user_id) {

        $index_lock_filename =~ s/\.xml$/.$lock_user_id.$section_id.xml/;

    # the edited file is locke dou tby thus user, or it is a fresh edit
    } elsif ($this->hasPermission('edit')) {

        $index_lock_filename =~ s/\.xml$/.$user_id.$section_id.xml/;

    }

    $this->_debug('index_lock_filename(1)', $index_lock_filename);

    # if this locked out version exists, then we use that
    return $index_lock_filename;
}

# this is a file of the form "index.section_id.xml"
sub _build_section_index_lock_filename {

    my $this = shift;
    my $args = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO }; # used for dirname

    my $section_id = $this->_get_section_id($path_info);

    $path_info = (dirname($path_info) ne '/' ? dirname($path_info) : '').$this->getEditFilename('index');
    my $section_index_lock_filename = $this->getEditFilename('root').$path_info;
    $section_index_lock_filename =~ s/\.xml$/.$section_id.xml/;

    $this->_debug('section_index_lock_filename', $section_index_lock_filename);

    return $section_index_lock_filename;
}


sub _build_index_template_filename {

    my $this = shift;
    my $args = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO };

    return dirname($this->_build_edit_filename({ USER_ID => $user_id, PATH_INFO => $path_info })).$this->getEditFilename('template-index');
}

sub _build_index_html_filename {

    my $this = shift;
    my $args = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO };

    # use lock_filename with a different path_info
    my $index_html_filename = $this->_build_index_lock_filename({ USER_ID => $user_id, PATH_INFO => $path_info });
    $index_html_filename =~ s/\.xml$/.html/;

    $this->_debug('index_html_filename(9)', $index_html_filename);

    # and back the the caller
    return $index_html_filename;

}

sub _build_section_index_html_filename {

    my $this = shift;
    my $args = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO };

    # use lock_filename with a different path_info
    my $section_index_html_filename = $this->_build_section_index_lock_filename({ USER_ID => $user_id, PATH_INFO => $path_info });
    $section_index_html_filename =~ s/\.xml$/.html/;

    $this->_debug('section_index_html_filename', $section_index_html_filename);

    # and back the the caller
    return $section_index_html_filename;
}

sub _build_index_browse_filename {

    my $this = shift;
    my $args = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO };

    my $index_browse_filename;
    my $section_id = $this->_get_section_id($path_info);

    # may be one of a few posibilities
    #   (1) if they have edit.section permission and index.section_id.html
    #   exists then we use that
    #
    #   (2) if they have a index.user_id.section.html available then use that
    #
    #   (3) use the standard browse index.html file

    # (1) check for edit.section permission
    my $section_index_html_filename = $this->_build_section_index_html_filename({ USER_ID => $user_id, PATH_INFO => $path_info });
    return $section_index_html_filename if ($this->hasPermission('edit.section') && -w $section_index_html_filename);

    # see what the same locked filename would be
    $index_browse_filename = $this->getEditFilename('root').(dirname($path_info) ne '/' ? dirname($path_info) : '').$this->getEditFilename('index');
    $index_browse_filename =~ s/\.xml$/.$user_id.$section_id.html/;

    $this->_debug('index_browse_filename', $index_browse_filename);

    # (2) check this file exists, as it may not if nothing has been edited yet
    return $index_browse_filename if -w $index_browse_filename;

    # (3) so we just rely on the normal html file in browse-land
    $index_browse_filename = $this->getBrowseFilename('root').(dirname($path_info) ne '/' ? dirname($path_info) : '').$this->getBrowseFilename('index');

    $this->_debug('index_browse_filename', $index_browse_filename);

    return $index_browse_filename;
}

sub isIndexFile {

    my $this = shift;
    my $path_info = shift;

    $path_info = $this->{ PATH_INFO } unless defined $path_info;

    # convert path_info to xml filename
    $path_info =~ s/\.html$/.xml/;

    $this->_debug('path_info(8)', $path_info);
    $this->_debug('basename(path_info)', basename($path_info));
    $this->_debug('basename($this->getEditFilename(index))', basename($this->getEditFilename('index')));

    # now check if this matches
    return basename($path_info) eq basename($this->getEditFilename('index'));
}

sub _parse_index_document {

    my $DOMTree = shift;
    my $content = shift;    # might not be supplied
    my $delete_flag = shift;     # might not be supplied

    my $document;

    Kumera::_debug('DOMTree->getNodeName()', $DOMTree->getNodeName());
    Kumera::_debug('delete_flag', $delete_flag);

    # there may, or may not be a DOCTYPE, which we ignore
    my $docNodes = $DOMTree->getChildNodes();
    for (my $i = 0 ; $i < $docNodes->getLength() ; $i++) {

        my $docNode = $docNodes->item($i);

        # look for the document element
        if ($docNode->getNodeName() eq 'Kumera:document') {

            # parse the content section, should only ever
            $document = _parse_entries($DOMTree, $docNode, $content, $delete_flag);
        }
    }

    # and return a reference to the hash
    return $document;
}

sub _parse_entries {

    my $DOMTree = shift;
    my $documentTree = shift;
    my $content = shift;
    my $delete_flag = shift;

    my @entries;
    my $a_match = 0;
    my $exists;

    my @delete_list;

    Kumera::_debug('documentTree->getNodeNames', $documentTree->getNodeName());
    Kumera::_debug('delete_flag', $delete_flag);

    my $entryNodes = $documentTree->getChildNodes();

    for (my $i = 0 ; $i < $entryNodes->getLength() ; $i++) {

        my $entryNode = $entryNodes->item($i);

        # check it is an entry node
        if ($entryNode->getNodeName() eq 'entry') {

            # get the url of this entry, and see if the content matches it
            my $url = $entryNode->getAttribute('url');

            # now either we have content
            # (1) that we already have here or
            # (2) this is new content
            # (3) content to be deleted

            # check for (1), by seeing if there is a url that matches
            $exists = grep { $_->[0] eq 'url' && $_->[2] eq $url } @{ $content } if defined $content;

            Kumera::_debug('url', $url);
            Kumera::_debug('exists', $exists);

            # if exists is non-zero then we have a match for this entry
            # and so we pump it through the system, if no match occurs
            # at all, then w ehave a new entry that we need to deal with
            $a_match = $a_match || $exists;

            # if it is there, and we are to delete it then do so
            if ($exists && defined $delete_flag && $delete_flag) {

                Kumera::_debug('delete_list', $url);

                # keep for now, so we can process outside of loop
                push @delete_list, $entryNode;

            } else {
                # push it through ...
                push @entries, _parse_content($entryNode, $exists ? $content : undef);
            }
        }
    }

    # we may be here, and the content was not put in, which means
    # that this is new content, so we need to create a new entry
    if (defined $content && !$a_match && !$delete_flag) {

        # the url of this content
        my ($url) = grep { $_->[0] eq 'url' } @{ $content };

        Kumera::_debug('url', $url);

        # first argument seem to be a XML::DOM::Document argument
        my $entryNode = new XML::DOM::Element($DOMTree, 'entry');

        # put the url attribute into the element
        $entryNode->setAttribute('url', $url->[2]);

        # now make children nodes for the content elements
        foreach my $this_content (@{ $content }) {

            next unless grep { $_ =~ $this_content->[0] } qw/order title abstract url/;

            # create the content element
            my $contentNode = new XML::DOM::Element($DOMTree, 'content');

            # set the name (we don't use type here)
            $contentNode->setAttribute('name', $this_content->[0]);

            # set the text
            $contentNode->addText($this_content->[2]);

            # and make this a child of the entryNode
            $entryNode->appendChild($contentNode);
        }

        # and put this child at the end (still need to sort them by order)
        $documentTree->appendChild($entryNode);

        # parse the content into the entries list as well
        push @entries, _parse_content($entryNode);

    # go through the delete list now
    } elsif ($delete_flag) {

        foreach my $node (@delete_list) {

            Kumera::_debug('node', $node);

            # this only works here because entry nodes are top level
            # if this changes then the parent of an entry node should be used
            $documentTree->removeChild($node);

        }

    }

    return \@entries;
}

sub _parse_content {

    my $documentTree = shift;
    my $content = shift;    # might not be supplied

    my %contentArray;

    Kumera::_debug('documentTree->getNodeNames', $documentTree->getNodeName());

    # document is made up of a data section
    my $contentNodes = $documentTree->getChildNodes();
    for (my $i = 0 ; $i < $contentNodes->getLength() ; $i++) {

        my $contentNode = $contentNodes->item($i);

        # check if it is a data node
        if ($contentNode->getNodeName() eq 'content') {

            my $name = $contentNode->getAttribute('name');

            Kumera::_debug('name', $name);

            # look for the entry in content that matches this name, and stuff it in
            if (defined $content) {

                my ($this_content) = grep { Kumera::_debug('$_->[0]', $_->[0]); $_->[0] eq $name } @{ $content };

                if (defined $this_content) {

                    Kumera::_debug('this_content->[0]', $this_content->[0]);

                    # if we have content supplied, and it matched this one, put it in
                    $contentNode->getFirstChild() ? $contentNode->getFirstChild()->setData($this_content->[2]) : $contentNode->addText($this_content->[2]);
                }
            }

            # the first child will be a text element, which we get the data for
            $contentArray{ uc $name } = defined $contentNode->getFirstChild() ? $contentNode->getFirstChild()->getData() : '';

        }
    }

    return \%contentArray;
}



# this will build both the user's index file and the editor's index file
sub _rebuild_indexfile {

    my $this = shift;
    my $args = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO };
    my $delete_flag = defined $args->{ DELETE } && $args->{ DELETE } ? 1 : 0;

    $args = { USER_ID => $user_id, PATH_INFO => $path_info, DELETE => $delete_flag};

    # and the index_filename for building indexes
    my $index_edit_filename = $this->_build_index_edit_filename($args);

    # and the index_filename for building indexes
    my $index_lock_filename = $this->_build_index_lock_filename($args);

    # and the html file that index will be written to
    my $index_html_filename = $this->_build_index_html_filename($args);

    # and the two index file names for section editors, etc
    my $section_index_lock_filename = $this->_build_section_index_lock_filename($args);

    # and the html file that section index will be written to
    my $section_index_html_filename = $this->_build_section_index_html_filename($args);

    $this->_debug('index_edit_filename(0)', $index_edit_filename);
    $this->_debug('index_lock_filename', $index_lock_filename);
    $this->_debug('index_html_filename', $index_html_filename);

    $this->_debug('section_index_lock_filename', $section_index_lock_filename);
    $this->_debug('section_index_html_filename', $section_index_html_filename);

    # this should not be done if they have edit.section permission, and this is their
    # own locked out file (so if they edited someone else's file then write that users
    # index_lock_filename, etc

    my $lock_user_id = $this->locked_out_by;
    $this->_debug('user_id', $this->{ USER_ID });
    $this->_debug('lock_user_id', $lock_user_id);

    # this one builds the index file for the user
    unless ($this->hasPermission('edit.section') && $this->{ USER_ID } eq $lock_user_id) {
        $this->_build_indexfile($args, $index_edit_filename, $index_lock_filename, $index_html_filename, $this->{ EDIT_DATA });
    }

    # and then we build the index file for the section
    $this->_build_indexfile($args, $index_edit_filename, $section_index_lock_filename, $section_index_html_filename, $this->{ EDIT_DATA });
}

sub _build_indexfile {

    my $this = shift;
    my $args = shift;
    my $index_edit_filename = shift;
    my $index_lock_filename = shift;
    my $index_html_filename = shift;
    my $edit_data = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO };
    my $delete_flag = defined $args->{ DELETE } && $args->{ DELETE } ? 1 : 0;

    $args = { USER_ID => $user_id, PATH_INFO => $path_info, DELETE => $delete_flag };

    # first, read in the index file if not already there
    my $DOMTree;

    $this->_debug('index_edit_filename', $index_edit_filename);
    $this->_debug('index_lock_filename', $index_lock_filename);
    $this->_debug('index_html_filename', $index_html_filename);

    # make the index_lock_filename file first, unless it already exists
    if (!-w $index_lock_filename) {
        die "Cannot create lock file: $!" unless copy $index_edit_filename, $index_lock_filename;
    }

    $DOMTree = new XML::DOM::Parser->parsefile($index_lock_filename);

    # now go through the DOMTree and push in the new information if supplied
    my $index_data = _parse_index_document($DOMTree, $edit_data, $delete_flag);

    # and now write the index DOM tree out as an xml file
    # for this user only ... i.e. as a locked file style
    $DOMTree->printToFile($index_lock_filename);
    chmod $Kumera::file_perms, $index_lock_filename;

    # and now write it out as a html file (do this in a different section)
    $this->_make_html_indexfile($args, $index_html_filename, $index_data);
}

sub _make_html_indexfile {

    my $this = shift;
    my $args = shift;
    my $index_html_filename = shift;
    my $index_data = shift;

    my $user_id = $args->{ USER_ID } || $this->{ USER_ID };
    my $path_info = $args->{ PATH_INFO } || $this->{ PATH_INFO };

    $args = { USER_ID => $user_id, PATH_INFO => $path_info };

    # and the template that INDEX will be using
    $this->{ INDEX_TEMPLATE_FILENAME } = $this->_build_index_template_filename($args);
    $this->_debug('this->{ INDEX_TEMPLATE_FILENAME }', $this->{ INDEX_TEMPLATE_FILENAME });

    # read in the template file
    my $input = new IO::File($this->{ INDEX_TEMPLATE_FILENAME });
    $input->untaint;
    my $template = join '', $input->getlines;
    $input->close;

    # build the list that template expects, sorted (hmmm)
    my @lines = sort { $a->{ ORDER } <=> $b->{ ORDER } } @{ $index_data };

    $this->_debug('lines', join ',', @lines);
    $this->_debug('index_html_filename', $index_html_filename);

    # now search for the [[ ]] bits to parse and execute
    $template =~ s/\[\[(.*?)\]\]/$1/eesg;

    # and save the html away to disk
    my $output = new IO::File($index_html_filename, O_CREAT|O_TRUNC|O_WRONLY, $Kumera::file_perms);
    $output->write($template, length $template);
    $output->close;

    # because IO::File doesn't seem to do it right ...
    chmod $Kumera::file_perms, $index_html_filename;
}

sub _rebuild_index_files {

    my $this = shift;
    my $file_list = shift;
    my $delete_flag = shift;

    # loop through the complete list of files
    foreach $path_info (@{ $file_list }) {

        # the user_id is used for working out which html template to use, which is
        # not a problem at this point since that really comes from path_info

        my $index_edit_filename = $this->_build_index_edit_filename({ USER_ID => $this->{ USER_ID }, PATH_INFO => $path_info });
        my $index_lock_filename = $index_edit_filename;
        my $index_html_filename = $this->getBrowseFilename('root').(dirname($path_info) ne '/' ? dirname($path_info) : '').$this->getBrowseFilename('index');

        $this->_debug('path_info', $path_info);
        $this->_debug('index_edit_filename(8)', $index_edit_filename);
        $this->_debug('index_html_filename(8)', $index_html_filename);

        # this is to work out the name of the index file (might be locked out form)
        my $edit_filename = $this->_build_edit_filename({ USER_ID => $this->{ USER_ID }, PATH_INFO => $path_info });
        my $lock_filename = $this->_build_lock_filename({ USER_ID => $this->locked_out_by({ PATH_INFO => $path_info }), PATH_INFO => $path_info });

        $this->_debug('edit_filename(2)', $edit_filename);
        $this->_debug('lock_filename(2)', $lock_filename);

        my $parser = new XML::DOM::Parser;

        # parse the one that exists, need to think about this better!
        my $edit_data = _parse_edit_document($parser->parsefile(-w $edit_filename ? $edit_filename : $lock_filename));

        $this->_debug('delete_flag', $delete_flag);

        # next create the index.xml and htdocs's index.html file
        $this->_build_indexfile
        (
            { USER_ID => $this->{ USER_ID }, PATH_INFO => $path_info, DELETE => $delete_flag },
            $index_edit_filename,
            $index_lock_filename,
            $index_html_filename,
            $edit_data,
        );

        # and then rebuild all the index.user_id.section_id.xml
        # files for each user that has one

        # build the source index_edit_filename (doesn't need $user_id)
        $index_edit_filename = $this->_build_index_edit_filename({ USER_ID => $this->{ USER_ID }, PATH_INFO => $path_info });
        $this->_debug('index_edit_filename(9)', $index_edit_filename);

        my $section_id = $this->_get_section_id($path_info);

        # read in the new edit_data
        $edit_data = _parse_edit_document($parser->parsefile(-w $edit_filename ? $edit_filename : $lock_filename));

        foreach my $user_id ($this->allUsers) {

            # build the destination filename
            $index_lock_filename = $index_edit_filename;
            $index_lock_filename =~ s/\.xml$/.$user_id.$section_id.xml/;

            # and the html filename
            ($index_html_filename = $index_lock_filename) =~ s/\.xml$/.html/;

            $this->_debug('path_info', $path_info);
            $this->_debug('user_id(9)', $user_id);
            $this->_debug('index_edit_filename', $index_edit_filename);
            $this->_debug('index_lock_filename', $index_lock_filename);
            $this->_debug('index_html_filename', $index_html_filename);
            $this->_debug('edit_filename(3)', $edit_filename);

            # don't create an index file if one doesn't exist (this will include
            # section editors, who always have an up to date index file
            next unless -w $index_lock_filename;

            # and build this index file
            $this->_build_indexfile
            (
                { USER_ID => $user_id, PATH_INFO => $path_info, DELETE => $delete_flag },
                $index_edit_filename,
                $index_lock_filename,
                $index_html_filename,
                $edit_data,
            );
        }

        # if this is a delete, we also need to rebuild the section
        # index file
        if ($delete_flag) {

            # and the two index file names for section editors, etc
            my $section_index_lock_filename = $this->_build_section_index_lock_filename({ PATH_INFO => $path_info });

            # and the html file that section index will be written to
            my $section_index_html_filename = $this->_build_section_index_html_filename({ PATH_INFO => $path_info });

            # and rebuild this one too
            $this->_build_indexfile
            (
                { PATH_INFO => $path_info, DELETE => $delete_flag },
                $index_edit_filename,
                $section_index_lock_filename,
                $section_index_html_filename,
                $edit_data
            );

        }
    }

}


####
##
## stuff for parsing the edit files
##
####


sub _parse_edit_document {

    my $DOMTree = shift;
    my $content = shift;    # might not be supplied

    my $document;

    Kumera::_debug('DOMTree->getNodeName()', $DOMTree->getNodeName());

    # there may, or may not be a DOCTYPE, which we ignore
    my $docNodes = $DOMTree->getChildNodes();
    for (my $i = 0 ; $i < $docNodes->getLength() ; $i++) {

        my $docNode = $docNodes->item($i);

        # look for the document element
        if ($docNode->getNodeName() eq 'Kumera:document') {

            # parse the content section, should only ever
            $document = _parse_edit_content($docNode, $content);
        }
    }

    # and return a reference to the hash
    return $document;
}


use constant NAME => 0;
use constant TYPE => 1;
use constant TEXT => 2;


sub _parse_edit_content {

    my $documentTree = shift;
    my $content = shift;    # might not be supplied

    my @contentArray;

    Kumera::_debug('documentTree->getNodeNames', $documentTree->getNodeName());

    # document is made up of a data section
    my $contentNodes = $documentTree->getChildNodes();
    for (my $i = 0 ; $i < $contentNodes->getLength() ; $i++) {

        my $contentNode = $contentNodes->item($i);

        # check if it is a data node
        if ($contentNode->getNodeName() eq 'content') {

            my $name = $contentNode->getAttribute('name');
            my $type = $contentNode->getAttribute('type');

            Kumera::_debug('content-name', $name);

            # if we have content supplied, and it matched this one, put it in
            ($contentNode->getFirstChild() ? $contentNode->getFirstChild()->setData($content->[1]) : $contentNode->addText($content->[1])) if (defined $content->[0] && $name eq $content->[0]);

            # the first child will be a text element, which we get the data for
            push @contentArray, [ $name, $type, defined $contentNode->getFirstChild() ? $contentNode->getFirstChild()->getData() : '' ];

            Kumera::_debug('contentArray', join ', ', @contentArray);
            Kumera::_debug('$#contentArray', $#contentArray);
            Kumera::_debug('contentArray['.$#contentArray.']', $contentArray[$#contentArray]);
            Kumera::_debug('contentArray['.$#contentArray.']->[NAME]', $contentArray[$#contentArray]->[NAME]);
            Kumera::_debug('contentArray['.$#contentArray.']->[TYPE]', $contentArray[$#contentArray]->[TYPE]);
        }
    }

    return \@contentArray;
}


1;
